import os
import sys
from importlib.metadata import metadata

import requests.exceptions
import typer
from chris.models import Pipeline, InvalidFilesResourceUrlException
from typing import Optional, List
import logging
from pathlib import Path

from caw.movedata import upload as cube_upload, download as cube_download
from caw.login import LoginManager
from caw.globals import DEFAULT_ADDRESS, DEFAULT_USERNAME, DEFAULT_PASSWORD
from caw.helpers import ClientPrecursor, run_pipeline

if 'CAW_DEBUG' in os.environ:
    logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


login_manager = LoginManager()
precursor = ClientPrecursor(login_manager)

app = typer.Typer(
    epilog='Examples and documentation at '
           'https://github.com/FNNDSC/caw#documentation'
)


def show_version(value: bool):
    """
    Print version.
    """
    if not value:
        return
    program_info = metadata(__package__)
    typer.echo(program_info['version'])
    raise typer.Exit()


@app.callback()
def main(
        address: str = typer.Option(DEFAULT_ADDRESS, '--address', '-a', envvar='CHRIS_URL'),
        username: str = typer.Option(DEFAULT_USERNAME, '--username', '-u', envvar='CHRIS_USERNAME'),
        password: str = typer.Option(DEFAULT_PASSWORD, '--password', '-p', envvar='CHRIS_PASSWORD'),
        version: Optional[bool] = typer.Option(None, '--version', '-V',
                                               callback=show_version, is_eager=True,
                                               help='Print version.')
):
    """
    A command line ChRIS client for pipeline execution and data management.
    """
    global precursor
    precursor.address = address
    precursor.username = username
    precursor.password = password


@app.command()
def login(read_pass: bool = typer.Option(False, '--password-stdin', help='Take the password from stdin')):
    """
    Login to ChRIS.
    """
    if precursor.username == DEFAULT_USERNAME:
        precursor.username = typer.prompt('username')
    if read_pass:
        precursor.password = ('\n'.join(sys.stdin)).rstrip('\n')
    elif precursor.password == DEFAULT_PASSWORD:
        precursor.password = typer.prompt('password', hide_input=True)

    client = precursor()
    login_manager.login(client.addr, client.token)


@app.command()
def logout():
    """
    Remove your login credentials.
    """
    if precursor.address == DEFAULT_ADDRESS:
        login_manager.logout()
    else:
        login_manager.logout(precursor.address)


@app.command()
def search(name: str = typer.Argument('', help='name of pipeline to search for')):
    """
    Search for pipelines that are saved in ChRIS.
    """
    client = precursor()
    for search_result in client.search_pipelines(name):
        typer.echo(f'{search_result.url:<60}{typer.style(search_result.name, bold=True)}')


@app.command()
def pipeline(name: str = typer.Argument(..., help='Name of pipeline to run.'),
             target: str = typer.Option(..., help='Plugin instance ID or URL.')):
    """
    Run a pipeline on an existing feed.
    """
    client = precursor()
    plugin_instance = client.get_plugin_instance(target)
    chris_pipeline = client.get_pipeline(name)
    run_pipeline(chris_pipeline=chris_pipeline, plugin_instance=plugin_instance)


@app.command()
def upload(
        threads: int = typer.Option(4, '--threads', '-t', help='Number of threads to use for file upload.'),
        create_feed: bool = typer.Option(True, help='Run pl-dircopy on the newly uploaded files.'),
        name: str = typer.Option('', '--name', '-n', help='Name of the feed.'),
        description: str = typer.Option('', '--description', '-d', help='Description of the feed.'),
        pipeline_name: str = typer.Option('', '--pipeline', '-p', help='Name of pipeline to run on the data.'),
        files: List[Path] = typer.Argument(..., help='Files to upload. '
                                                     'Folder upload is supported, but directories are destructured.')
):
    """
    Upload local files and run pl-dircopy.
    """
    client = precursor()
    chris_pipeline: Optional[Pipeline] = None
    if pipeline_name:
        chris_pipeline = client.get_pipeline(pipeline_name)

    try:
        swift_path = cube_upload(client=client, files=files, upload_threads=threads)
    except requests.exceptions.RequestException as e:
        logger.debug('RequestException: %s\n%s', str(e), e.response.text)
        typer.secho('Upload unsuccessful', fg=typer.colors.RED, err=True)
        raise typer.Abort()

    if not create_feed:
        raise typer.Exit()

    dircopy_instance = client.run('pl-dircopy', params={'dir': swift_path})
    if name:
        dircopy_instance.get_feed().set_name(name)
    if description:
        dircopy_instance.get_feed().set_description(description)

    if chris_pipeline:
        run_pipeline(chris_pipeline=chris_pipeline, plugin_instance=dircopy_instance)
    typer.echo(dircopy_instance.feed)


@app.command()
def download(
        threads: int = typer.Option(4, '--threads', '-t', help='Number of concurrent downloads.'),
        url: str = typer.Argument(..., help='ChRIS files API resource URL'),
        destination: Path = typer.Argument(..., help='Location on host where to save downloaded files.')
):
    """
    Download everything from a ChRIS url.
    """
    client = precursor()
    try:
        cube_download(client=client, url=url, destination=destination, threads=threads)
    except InvalidFilesResourceUrlException as e:
        typer.secho(str(e), fg=typer.colors.RED, err=True)
        raise typer.Abort()


if __name__ == '__main__':
    app()



"""
Session token storage mechanisms.

If available, the desktop keyring is used. Else a session token
(not the account password) will be stored in plaintext.
"""
import abc
import json
from abc import ABC
from typing import Optional, Callable
import typer
from pathlib import Path

import logging

logger = logging.getLogger(__name__)

use_keyring = True
try:
    import keyring
    logger.info('Using keyring')
except ModuleNotFoundError:
    use_keyring = False
    logger.info('Using plaintext')


class NotLoggedInError(Exception):
    pass


class AbstractSecretStore(ABC):
    """
    A key-value data structure which hopefully stores its values securely.
    It can read and write to the application's settings, though saving of those
    settings is not its responsibility.

    The application's settings are not necessarily stored securely. Instead, it
    is the responsibility of this AbstractSecretStore to store secrets securely
    somewhere else (optionally using the application's settings as pointers).
    """

    def __init__(self, context: dict):
        """
        :param context: application settings which are expected to be saved
                        after the invocation of this object's methods
        """
        self.context = context

    @abc.abstractmethod
    def get(self, address: str) -> str:
        ...

    @abc.abstractmethod
    def set(self, address: str, token: str):
        ...

    @abc.abstractmethod
    def clear(self, address: str):
        ...


class KeyringSecretStore(AbstractSecretStore):
    """
    Secure token storage using the host desktop environment's login keyring.
    """

    __SPACE = 'org.chrisproject.caw'

    def get(self, address: str) -> str:
        token = keyring.get_password(self.__SPACE, address)
        if not token:
            raise NotLoggedInError()
        return token

    def set(self, address: str, token: str):
        keyring.set_password(self.__SPACE, address, token)

    def clear(self, address: str):
        keyring.delete_password(self.__SPACE, address)


class FallbackSecretStore(AbstractSecretStore):
    """
    Insecure token storage in plaintext.
    This implementation should only be used when the keyring is not available.
    """
    def __init__(self, context: dict):
        super().__init__(context)
        if 'secrets' not in context:
            context['secrets'] = {}

    def get(self, address: str) -> str:
        if address not in self.context['secrets']:
            raise NotLoggedInError()
        return self.context['secrets'][address]

    def set(self, address: str, token: str):
        self.context['secrets'][address] = token

    def clear(self, address: str):
        del self.context['secrets'][address]


class LoginManager:
    """
    A wrapper around an AbstractSecretStore which handles saving of configurations.

    The configuration file is saved in ~/.config/caw after each method call.
    """

    def __init__(self, PreferredSecretStore: Optional[Callable[[dict], AbstractSecretStore]] =
                 KeyringSecretStore if use_keyring else FallbackSecretStore,
                 config_file: Optional[Path] = Path(typer.get_app_dir('caw')) / 'login.json'):
        # new empty configuration on first run
        if not config_file.exists():
            appdir = Path(typer.get_app_dir('caw'))
            if config_file.parent == appdir:
                appdir.mkdir(parents=True)
            config_file.write_text('{}')

        self.__savefile = config_file
        with self.__savefile.open('r') as f:
            self.__store = PreferredSecretStore(json.load(f))

    def get_default_address(self) -> Optional[str]:
        if 'defaultAddress' not in self.__store.context:
            return None
        return self.__store.context['defaultAddress']

    def __default_address(self, address: Optional[str] = None):
        """
        Get the default address from the configuration if the address was not given.
        :param address: CUBE address
        :return: CUBE address
        """
        if address:
            return address
        default_address = self.get_default_address()
        if not default_address:
            raise NotLoggedInError('Default CUBE address has not yet been set.')
        return default_address

    def __write_config(self):
        """
        Save login configuration to disk.
        """
        with self.__savefile.open('w') as f:
            json.dump(self.__store.context, f)

    def get(self, address: Optional[str] = None):
        address = self.__default_address(address)
        return self.__store.get(address)

    def logout(self, address: Optional[str] = None):
        """
        Remove secret from storage. If the address is the default address, then remove the default address as well.
        :param address: CUBE address
        """
        if not address:
            address = self.__default_address(address)
            del self.__store.context['defaultAddress']

        self.__store.clear(address)
        self.__write_config()

    def login(self, address: str, token: str):
        self.__store.set(address, token)
        self.__store.context['defaultAddress'] = address
        self.__write_config()

        if not use_keyring and isinstance(self.__store, FallbackSecretStore):
            typer.secho(f'Login token was saved as plaintext in the file ', dim=True, err=True)
            typer.secho(str(self.__savefile), dim=True, err=True)
            typer.secho('For safer identification storage, please run: '
                        '\n\n\tcaw logout'
                        '\n\tpip install keyring'
                        f'\n\tcaw login --address {address}\n',  dim=True, err=True)



from os import path
import requests
from typing import Optional, Set, Union

from chris.models import PluginInstance, Plugin, Pipeline, UploadedFiles


class ChrisClientError(Exception):
    pass


class ChrisIncorrectLoginError(ChrisClientError):
    pass


class ChrisResourceNotFoundError(ChrisClientError):
    pass


class PluginNotFoundError(ChrisResourceNotFoundError):
    pass


class PipelineNotFoundError(ChrisResourceNotFoundError):
    pass


class ChrisClient:
    def __init__(self, address: str, username: Optional[str] = None, password: Optional[str] = None,
                 token: Optional[str] = None):
        """
        Log into ChRIS.
        :param address: CUBE address
        :param username: account username
        :param password: account password
        :param token: use token authorization, takes priority over basic authorization
        """
        if not address.endswith('/api/v1/'):
            raise ValueError('Address of CUBE must end with "/api/v1/"')
        self.addr = address
        self.search_addr_plugins = address + 'plugins/search/'
        self.search_addr_plugins_instances = address + 'plugins/instances/search/'
        self.search_addr_pipelines = address + 'pipelins/search/'

        self._s = requests.Session()
        self._s.headers.update({'Accept': 'application/json'})

        if not token:
            if not username or not password:
                raise ChrisIncorrectLoginError('Username and password are required.')

            auth_url = self.addr + 'auth-token/'
            login = self._s.post(auth_url, json={
                'username': username,
                'password': password
            })
            if login.status_code == 400:
                res = login.json()
                raise ChrisIncorrectLoginError(res['non_field_errors'][0] if 'non_field_errors' in res else login.text)
            login.raise_for_status()
            token = login.json()['token']

        self._s.headers.update({
            'Content-Type': 'application/vnd.collection+json',
            'Authorization': 'Token ' + token
        })
        self.token = token
        """
        HTTP basic authentication token.
        """

        res = self._s.get(address)
        if res.status_code == 401:
            data = res.json()
            raise ChrisIncorrectLoginError(data['detail'] if 'detail' in data else res.text)
        if res.status_code != 200:
            raise ChrisClientError(f'CUBE response status code was {res.status_code}.')
        res.raise_for_status()
        data = res.json()
        if 'collection_links' not in data or 'uploadedfiles' not in data['collection_links']:
            raise ChrisClientError(f'Unexpected CUBE response: {res.text}')
        self.collection_links = data['collection_links']

        res = self._s.get(self.collection_links['user'])
        res.raise_for_status()
        data = res.json()
        self.username = data['username']
        """
        The ChRIS user's username.
        """

    def upload(self, file_path: str, upload_folder: str):
        """
        Upload a local file into ChRIS backend Swift storage.
        :param file_path: local file path
        :param upload_folder: path in Swift where to upload to
        :return: response
        """
        bname = path.basename(file_path)
        upload_path = path.join(upload_folder, bname)

        with open(file_path, 'rb') as file_object:
            files = {
                'upload_path': (None, upload_path),
                'fname': (bname, file_object)
            }
            res = self._s.post(
                self.collection_links['uploadedfiles'],
                files=files,
                headers={
                    'Accept': 'application/vnd.collection+json',
                    'Content-Type': None
                }
            )
        res.raise_for_status()
        return res.json()

    def _url2plugin(self, url):
        res = self._s.get(url)
        res.raise_for_status()
        return Plugin(**res.json(), session=self._s)

    def get_plugin(self, name_exact='', version='', url='') -> Plugin:
        """
        Get a single plugin, either searching for it by its exact name, or by URL.
        :param name_exact: name of plugin
        :param version: (optional) version of plugin
        :param url: (alternative to name_exact) url of plugin
        :return:
        """
        if name_exact:
            search = self.search_plugin(name_exact, version)
            return search.pop()
        elif url:
            return self._url2plugin(url)
        else:
            raise ValueError('Must give either plugin name or url')

    def search_plugin(self, name_exact: str, version: '') -> Set[Plugin]:
        payload = {
            'name_exact': name_exact
        }
        if version:
            payload['version'] = version
        res = self._s.get(self.search_addr_plugins, params=payload)
        res.raise_for_status()
        data = res.json()
        if data['count'] < 1:
            raise PluginNotFoundError(name_exact)
        return set(Plugin(**pldata, session=self._s) for pldata in data['results'])

    def get_plugin_instance(self, id: Union[int, str]):
        """
        Get a plugin instance.
        :param id: Either a plugin instance ID or URL
        :return: plugin instance
        """
        res = self._s.get(id if '/' in id else f'{self.addr}plugins/instances/{id}/')
        res.raise_for_status()
        return PluginInstance(**res.json())

    def run(self, plugin_name='', plugin_url='', plugin: Optional[PluginInstance] = None,
            params: Optional[dict] = None) -> PluginInstance:
        """
        Create a plugin instance. Either procide a plugin object,
        or search for a plugin by name or URL.
        :param plugin: plugin to run
        :param plugin_name: name of plugin to run
        :param plugin_url: alternatively specify plugin URL
        :param params: plugin parameters as key-value pairs (not collection+json)
        :return:
        """
        if not plugin:
            plugin = self.get_plugin(name_exact=plugin_name, url=plugin_url)
        return plugin.create_instance(params)

    def search_uploadedfiles(self, fname='', fname_exact='') -> UploadedFiles:
        query = {
            'fname': fname,
            'fname_exact': fname_exact
        }
        qs = '&'.join([f'{k}={v}' for k, v in query.items() if v])
        url = f"{self.collection_links['uploadedfiles']}search/?{qs}"
        return self.get_uploadedfiles(url)

    def get_uploadedfiles(self, url: str) -> UploadedFiles:
        return UploadedFiles(url=url, session=self._s)

    def search_pipelines(self, name='') -> Set[Pipeline]:
        payload = {
            'name': name
        }
        res = self._s.get(self.collection_links['pipelines'] + 'search/', params=payload)
        res.raise_for_status()
        data = res.json()
        return set(Pipeline(**p, session=self._s) for p in data['results'])

    def get_pipeline(self, name: str) -> Pipeline:
        search = self.search_pipelines(name)
        if not search:
            raise PipelineNotFoundError(name)
        return search.pop()



"""
This file eases usage of the Python chris.client and Typer.
"""
import os
import typer

from chris.models import Pipeline, PluginInstance
from chris.client import ChrisClient, ChrisIncorrectLoginError, PipelineNotFoundError
from caw.globals import DEFAULT_ADDRESS, DEFAULT_PASSWORD
from caw.login import LoginManager, NotLoggedInError

import logging
logger = logging.getLogger(__name__)


def run_pipeline(chris_pipeline: Pipeline, plugin_instance: PluginInstance):
    """
    Helper to execute a pipeline with a progress bar.
    """
    with typer.progressbar(plugin_instance.append_pipeline(chris_pipeline),
                           length=len(chris_pipeline.pipings), label='Scheduling pipeline') as proto_pipeline:
        for _ in proto_pipeline:
            pass


class FriendlyClient(ChrisClient):
    """
    A ``ChrisClient`` which shows (less helpful) error messages instead of exceptions.
    """
    def get_pipeline(self, name: str) -> Pipeline:
        try:
            return super().get_pipeline(name)
        except PipelineNotFoundError:
            typer.secho(f'Pipeline not found: "{name}"', fg=typer.colors.RED, err=True)
            raise typer.Abort()


class ClientPrecursor:
    """
    A workaround so that the ChrisClient object's constructor, which attempts to use the login
    credentials, is called by the subcommand instead of the main callback.

    This is necessary to support the ``caw login`` subcommand.
    """
    def __init__(self, login_manager: LoginManager):
        self.login_manager = login_manager

        self.address = None
        self.username = None
        self.password = None
        self.token = None

    def __call__(self) -> ChrisClient:
        """
        Authenticate with ChRIS and construct the client object.

        Login strategy:
        1. First, use given credentials.
        2. If credentials not specified, use saved login.

        :return: client object
        """
        if not self.address:
            raise ValueError('Must specify CUBE address.')

        # check if previously logged in
        if self.address == DEFAULT_ADDRESS:
            saved_address = self.login_manager.get_default_address()
            if saved_address:
                self.address = saved_address
        logger.debug('CUBE address: %s', self.address)

        if self.password == DEFAULT_PASSWORD:  # assume password not specified
            try:
                self.token = self.login_manager.get(self.address)
            except NotLoggedInError:  # password not specified and not previously logged in
                if 'CHRIS_TESTING' not in os.environ:
                    typer.secho('Using defaults (set CHRIS_TESTING=y to suppress this message): '
                                f'{self.address}  {self.username}:{self.password}', dim=True, err=True)

        # If previously logged in, use token to authenticate.
        # Otherwise, use username and password combo (which may or may not be be the default chris:chris1234).
        try:
            if self.token:
                logger.debug('HTTP token: "%s"', self.token)
                return FriendlyClient(self.address, token=self.token)
            else:
                return FriendlyClient(self.address, username=self.username, password=self.password)
        except ChrisIncorrectLoginError as e:
            typer.secho(e.args[0], err=True)
            raise typer.Abort()
        except Exception:
            typer.secho('Connection error\n'
                        f'address:  {self.address}\n'
                        f'username: {self.username}', fg=typer.colors.RED, err=True)
            raise typer.Abort()



import os
from concurrent.futures import ThreadPoolExecutor
import datetime
from typing import List, Tuple
import typer
from chris.client import ChrisClient
from chris.models import UploadedFile
import logging
from pathlib import Path


def upload(client: ChrisClient, files: List[Path], parent_folder='', upload_threads=4):

    if parent_folder:
        upload_folder = f'{client.username}/uploads/{parent_folder}/{datetime.datetime.now().isoformat()}/'
    else:
        upload_folder = f'{client.username}/uploads/{datetime.datetime.now().isoformat()}/'

    input_files = []
    for mri in files:
        if mri.is_file():
            input_files.append(mri)
        elif mri.is_dir():
            if len(files) != 1:
                typer.secho(f'WARNING: contents of {mri} will be uploaded to the top-level, '
                            'i.e. directory structure is not preserved.', dim=True, err=True)
            input_files += [f.absolute().name for f in mri.rglob('*')]
        else:
            typer.secho(f'No such file or directory: {mri}', fg=typer.colors.RED, err=True)
            raise typer.Abort()

    with typer.progressbar(label='Uploading files', length=len(input_files)) as bar:
        def upload_file(input_file: str):
            client.upload(input_file, upload_folder)
            bar.update(1)

        with ThreadPoolExecutor(max_workers=upload_threads) as pool:
            uploads = pool.map(upload_file, input_files)

    # check for upload errors
    for upload_result in uploads:
        logging.debug(upload_result)

    typer.secho(f'Successfully uploaded {len(input_files)} files to "{upload_folder}"', fg=typer.colors.GREEN, err=True)
    return upload_folder


def download(client: ChrisClient, url: str, destination: Path, threads: 4):
    """
    Download all the files from a given ChRIS API url.
    :param client: ChRIS client
    :param url: any ChRIS file resource url, e.g.
                https://cube.chrisproject.org/api/v1/uploadedfiles/
                https://cube.chrisproject.org/api/v1/uploadedfiles/?fname=chris/uploads/today
                https://cube.chrisproject.org/api/v1/3/files/
    :param destination: folder on host where to download to
    :param threads: max number of concurrent downloads
    """
    if destination.is_file():
        typer.secho(f'Cannot download into {destination}: is a file', fg=typer.colors.RED, err=True)
        raise typer.Abort()

    def __calculate_target(remote_file: UploadedFile) -> Tuple[Path, UploadedFile]:
        """
        Decide on a download location for a file resource in ChRIS.
        Create the parent directory if needed.
        :param remote_file: file information from ChRIS
        :return: download location on host and that file
        """
        fname = remote_file.fname
        if fname.startswith('chris/'):
            fname = fname[6:]
        target = destination.joinpath(fname)
        os.makedirs(target.parent, exist_ok=True)
        return target, remote_file

    search = client.get_uploadedfiles(url)
    with typer.progressbar(search, length=len(search), label='Getting information') as progress:
        to_download = frozenset(__calculate_target(remote_file) for remote_file in progress)

    with typer.progressbar(length=len(to_download), label='Downloading files') as progress:
        def download_file(t: Tuple[Path, UploadedFile]) -> int:
            """
            Download file and move the progress bar
            :param t: tuple
            :return: downloaded file size
            """
            target, remote_file = t
            remote_file.download(target)
            progress.update(1)
            return target.stat().st_size

        with ThreadPoolExecutor(max_workers=threads) as pool:
            sizes = pool.map(download_file, to_download)

    total_size = sum(sizes)
    if total_size < 2e5:
        size = f'{total_size} bytes'
    elif total_size < 2e8:
        size = f'{total_size / 1e6:.4f} MB'
    else:
        size = f'{total_size / 1e9:.4f} GB'
    typer.secho(size, fg=typer.colors.GREEN, err=True)



import os
import requests
from datetime import datetime
from chris.util import collection_helper, PaginationNotImplementedException
from typing import Optional, Set, Union, Iterator
from collections.abc import Iterable
from pathlib import Path
from abc import ABC
from queue import Queue
import logging


PAGINATION_LIMIT = os.getenv('CAW_PAGINATION_LIMIT', 64)


class ConnectedResource(ABC):
    """
    An object returned from the CUBE API which can make further requests to the API.
    """
    def __init__(self, url: str, session: requests.Session):
        self.url = url
        self._s = session

    def __hash__(self):
        return hash(self.url)


class Feed(ConnectedResource):
    def __init__(self, feed_url, session: requests.Session):
        super().__init__(feed_url, session)

        res = self._s.get(feed_url).json()
        self._note_url = res['note']

    def set_name(self, name):
        payload = collection_helper({'name': name})
        res = self._s.put(self.url, json=payload)
        res.raise_for_status()
        return res.json()

    def set_description(self, description):
        payload = collection_helper({
            'title': 'Description',
            'content': description
        })
        res = self._s.put(self._note_url, json=payload)
        res.raise_for_status()
        return res.json()


class PluginInstance(ConnectedResource):
    def __init__(self, url: str, id: int, title: str, feed: str, session: requests.Session, **kwargs):
        super().__init__(url, session)
        self.id = id
        self.title = title
        self.feed = feed

    def get_feed(self):
        return Feed(feed_url=self.feed, session=self._s)

    def append_pipeline(self, pipeline: 'Pipeline'):
        """
        Run a pipeline as a generator of plugin instances. Every ``next()`` creates a plugin instance.
        :param pipeline: pipeline to run
        """

        instances_map = {
            'None': self.id
        }
        """
        Maps Piping IDs to newly created PluginInstance IDs.
        """

        for p in pipeline:
            params = {
                'previous_id': instances_map[str(p.previous_id)]
            }
            params.update(p.default_parameters)
            next_instance = p.plugin.create_instance(params)
            yield next_instance
            instances_map[str(p.id)] = next_instance.id


class Plugin(ConnectedResource):
    def __init__(self, id: int, name: str, version: str,
                 instances: str, url: str, session: requests.Session, **kwargs):
        super().__init__(url, session)
        self.id = id
        self.name = name
        self.version = version
        self.instances = instances

    def create_instance(self, params: dict = None) -> PluginInstance:
        payload = collection_helper(params)

        res = self._s.post(self.instances, json=payload)
        res.raise_for_status()
        return PluginInstance(**res.json(), session=self._s)


class Piping(Iterable):
    """
    A Piping is the information about a plugin's membership of a pipeline.
    It is a node of a directed acyclic graph representation of a pipeline.
    Edges are bidirectional. A Piping knows its parent and its children.
    """

    def __init__(self, id: int, pipeline: str, pipeline_id: int, plugin: str, plugin_id: int, url: str,
                 default_parameters: dict, session: requests.Session, previous: Optional[str],
                 previous_id: Optional[int] = None):
        self.id = id
        self.pipeline = pipeline
        self.pipeline_id = pipeline_id
        self.plugin_id = plugin_id
        self.url = url
        self.previous = previous
        self.previous_id = previous_id
        self.children: Set['Piping'] = set()
        """Graph edges to children"""
        self.parent: Optional['Piping'] = None
        """Graph edge to parent"""
        self.default_parameters = default_parameters

        res = session.get(plugin)
        res.raise_for_status()
        self.plugin = Plugin(**res.json(), session=session)

    def __hash__(self):
        return hash((self.id, self.pipeline_id))

    def add_child(self, child: 'Piping'):
        self.children.add(child)

    def _to_queue(self) -> Queue:
        """
        Convert the tree to a queue where the root is first, and dependency plugins appear
        before the plugins which depend on them.
        :return: queue for a breadth-first traversal over the graph
        """
        q = Queue()
        self._add_all_to_queue(q, self)
        return q

    @classmethod
    def _add_all_to_queue(cls, q: Queue, p: 'Piping'):
        q.put(p)
        for c in p.children:
            cls._add_all_to_queue(q, c)

    def __iter__(self) -> Iterator['Piping']:
        """
        Breadth-first graph traversal.
        """
        q = self._to_queue()
        while not q.empty():
            yield q.get_nowait()
            q.task_done()


class PipelineAssemblyException(Exception):
    """
    Pipeline JSON representation cannot be reassembled as a Piping DAG.
    """
    pass


class PipelineHasMultipleRootsException(PipelineAssemblyException):
    """
    Multiple pipings with 'previous': null were found in the pipeline JSON representation.
    """
    pass


class PipelineRootNotFoundException(PipelineAssemblyException):
    """
    No piping found in the pipelines JSON representation with 'previous': null.
    """
    pass


class Pipeline(ConnectedResource):
    def __init__(self, authors: str, description: str, name: str, plugin_pipings: str, default_parameters: str,
                 plugins: str, url: str, session: requests.Session, **kwargs):
        super().__init__(url, session)
        self.authors = authors
        self.description = description
        self.name = name
        self.plugin_pipings = plugin_pipings
        self.default_parameters = default_parameters
        self.plugins = plugins
        self.pipings = self._do_get(self.plugin_pipings)

    def _do_get(self, url):
        res = self._s.get(url, params={'limit': PAGINATION_LIMIT, 'offset': 0})
        res.raise_for_status()
        data = res.json()
        if data['next']:
            raise PaginationNotImplementedException()
        return data['results']

    def get_default_parameters(self):
        return self._do_get(self.default_parameters)

    def assemble(self) -> Piping:
        """
        Convert the responses from CUBE to a DAG with parent --> child relationships
        (whereas CUBE's response represents a pipeline via child --> parent relationships
        through the previous key) and couples parameter info with plugin info.
        :return: DAG
        """
        # collect all default parameters
        assembled_params = {}
        for param_info in self.get_default_parameters():
            i = param_info['plugin_piping_id']
            if i not in assembled_params:
                assembled_params[i] = {}
            assembled_params[i][param_info['param_name']] = param_info['value']

        pipings_map = {}
        root: Optional[Piping] = None

        # create DAG nodes
        for piping_info in self.pipings:
            i = piping_info['id']
            if i in assembled_params:
                params = assembled_params[i]
            else:
                params = {}
            piping = Piping(**piping_info, default_parameters=params, session=self._s)
            pipings_map[i] = piping

            if not piping.previous:
                if root:
                    raise PipelineHasMultipleRootsException()
                root = piping
        if not root:
            raise PipelineRootNotFoundException()

        # create bidirectional DAG edges
        for _, piping in pipings_map.items():
            i = piping.previous_id
            if not i:
                continue
            pipings_map[i].add_child(piping)
            piping.parent = pipings_map[i]

        return root

    def __len__(self) -> int:
        """
        :return: number of pipings
        """
        return len(self.pipings)

    def __iter__(self):
        return iter(self.assemble())


class UploadedFile(ConnectedResource):
    def __init__(self, creation_date: str, file_resource: str, fname: str, fsize: int, id: int,
                 url: str, session: requests.Session,
                 owner: Optional[str] = None,
                 feed_id: Optional[int] = None, plugin_inst: Optional[str] = None,
                 plugin_inst_id: Optional[int] = None):
        super().__init__(url, session)
        self.creation_date = datetime.fromisoformat(creation_date)
        self.file_resource = file_resource
        self.fname = fname
        self.fsize = fsize
        self.id = id
        self.ownder = owner
        self.feed_id = feed_id
        self.plugin_inst = plugin_inst
        self.plugin_inst_id = plugin_inst_id

    def download(self, destination: Union[Path, str], chunk_size=8192):
        with self._s.get(self.file_resource, stream=True, headers={'Accept': None}) as r:
            r.raise_for_status()
            with open(destination, 'wb') as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    f.write(chunk)


class InvalidFilesResourceUrlException(Exception):
    pass


class UploadedFiles(ConnectedResource, Iterable):
    """
    Lazy iterable over paginated response.
    """
    __logger = logging.getLogger('UploadedFiles')

    def __init__(self, url: str, session: requests.Session):
        super().__init__(url, session)
        if 'limit=' not in self.url:
            self.url += f"{'&' if '?' in self.url else '?'}limit={PAGINATION_LIMIT}"

        try:
            self._initial_data = self._do_get(self.url)
        except requests.exceptions.HTTPError as e:
            raise InvalidFilesResourceUrlException(f'{e.response.status_code} error getting {self.url}')
        # check given URL is a files collection resource
        if 'count' not in self._initial_data:
            raise InvalidFilesResourceUrlException(f'{self.url} does not look like a files collection resource.')
        if self._initial_data['count'] > 0:
            try:
                UploadedFile(session=self._s, **self._initial_data['results'][0])
            except KeyError:
                raise InvalidFilesResourceUrlException(f'{self.url} is not a files collection resource.')

    def __iter__(self) -> Iterator[UploadedFile]:
        data = self._initial_data  # first page
        if data['previous'] is not None:
            self.__logger.warning('%s is not the first page.', self.url)

        while data['next']:
            for fdata in data['results']:
                yield UploadedFile(**fdata, session=self._s)
            self.__logger.debug('next page: %s', data['next'])
            data = self._do_get(data['next'])  # next page
        for fdata in data['results']:  # last page
            yield UploadedFile(**fdata, session=self._s)

    def __len__(self):
        return self._initial_data['count']

    def _do_get(self, url):
        self.__logger.info('getting %s', url)
        res = self._s.get(url)
        res.raise_for_status()
        return res.json()
